from dataclasses import dataclass
import typing
from commonfate_provider import (
    provider,
    resources,
    tasks,
    schema,
    access,
    rpc,
)


@dataclass
class AWSLambdaRuntime:
    provider: provider.Provider
    name: typing.Optional[str] = None
    version: typing.Optional[str] = None
    publisher: typing.Optional[str] = None

    def handle(self, event, context):
        result = self._do_handle(event=event, context=context)
        if result is not None:
            return result.dict(by_alias=True)

    def _do_handle(self, event, context) -> typing.Optional[rpc.Result]:
        parsed = rpc.Event.parse_obj(event)
        event = parsed.__root__

        if isinstance(event, rpc.Grant):
            grant_result = access.call_access_func(
                type="grant",
                p=self.provider,
                data=event.data,
            )

            if grant_result is None:
                return rpc.Result(response=rpc.GrantResponse())  # empty response

            response = rpc.GrantResponse(
                access_instructions=grant_result.access_instructions,
                state=grant_result.state,
            )

            return rpc.Result(response=response)

        elif isinstance(event, rpc.Revoke):
            return access.call_access_func(
                type="revoke",
                p=self.provider,
                data=event.data,
            )

        if isinstance(event, rpc.Describe):
            # Describe returns the configuration of the provider including the current status.
            provider = {
                "publisher": self.publisher,
                "name": self.name,
                "version": self.version,
            }
            config = self.provider._safe_config
            diagnostics = self.provider.diagnostics.export_logs()
            healthy = self.provider.diagnostics.has_no_errors()
            provider_schema = schema.export_schema().dict(
                exclude_none=True, by_alias=True
            )

            response = rpc.DescribeResponse(
                config=config,
                diagnostics=diagnostics,
                healthy=healthy,
                provider=provider,
                schema=provider_schema,
            )

            return rpc.Result(response=response)

        elif isinstance(event, rpc.Load):
            resources._reset()
            tasks._reset()
            tasks._execute(
                provider=self.provider, task=event.data.task, ctx=event.data.ctx
            )
            # find the resources and pending tasks, and return them
            found = resources.get()
            pending_tasks = tasks.get()

            response = rpc.LoadResponse(
                resources=[r.export_json() for r in found],
                tasks=[t.json() for t in pending_tasks],
            )

            return rpc.Result(response=response)

        else:
            raise Exception(f"unhandled event type")



import typing

from pydantic import BaseModel, Field


class AccessRequest(BaseModel):
    id: str


class GrantData(BaseModel):
    class Target(BaseModel):
        arguments: typing.Dict[str, str]
        kind: str

    subject: str
    target: Target
    state: typing.Optional[dict] = None
    request: typing.Optional[AccessRequest] = None


class Grant(BaseModel):
    type: typing.Literal["grant"]
    data: GrantData


class Revoke(BaseModel):
    type: typing.Literal["revoke"]
    data: GrantData


class Describe(BaseModel):
    type: typing.Literal["describe"]


class Load(BaseModel):
    class Data(BaseModel):
        task: str
        """the resource loader function ID to run"""
        ctx: typing.Optional[dict] = {}
        """context information for the task"""

    type: typing.Literal["load"]
    data: Data


class Event(BaseModel):
    __root__: typing.Union[Grant, Revoke, Load, Describe] = Field(
        ..., discriminator="type"
    )


class DescribeResponse(BaseModel):
    provider: dict
    config: dict
    diagnostics: typing.List[dict]
    healthy: bool
    provider_schema: dict = Field(alias="schema")


class LoadResponse(BaseModel):
    resources: typing.List[dict]
    tasks: typing.List[dict]


class GrantResponse(BaseModel):
    access_instructions: typing.Optional[str] = None
    state: typing.Optional[dict] = None


class Result(BaseModel):
    response: typing.Union[DescribeResponse, LoadResponse, GrantResponse]



from dataclasses import dataclass
from pydantic import BaseModel
import pytest
from commonfate_provider import access, namespace, provider, rpc


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


def test_target():
    @access.target()
    class ExampleTarget:
        pass

    targets = namespace.get_target_classes()

    assert targets["ExampleTarget"].cls == ExampleTarget


def test_target_with_kind():
    @access.target(kind="SomethingElse")
    class ExampleTarget:
        pass

    targets = namespace.get_target_classes()

    assert targets["SomethingElse"].cls == ExampleTarget


def test_call_revoke_works():
    class Provider(provider.Provider):
        pass

    @access.target()
    class ExampleTarget:
        pass

    class State(BaseModel):
        my_val: str

    got_var = None

    @access.revoke()
    def revoke(p: Provider, subject: str, target: ExampleTarget, state: State):
        nonlocal got_var
        got_var = state.my_val

    access.call_access_func(
        type="revoke",
        p=Provider(),
        data=rpc.GrantData(
            subject="test",
            target=rpc.GrantData.Target(
                kind="ExampleTarget",
                arguments={},
            ),
            state={"my_val": "test"},
        ),
    )

    # this should be set if the revoke function
    # was called with the correct state variable.
    assert got_var == "test"


def test_call_revoke_works_with_no_context():
    class Provider(provider.Provider):
        pass

    @access.target()
    class ExampleTarget:
        pass

    @access.revoke()
    def revoke(p: Provider, subject: str, target: ExampleTarget):
        pass

    access.call_access_func(
        type="revoke",
        p=Provider(),
        data=rpc.GrantData(
            subject="test",
            target=rpc.GrantData.Target(
                kind="ExampleTarget",
                arguments={},
            ),
            state=None,
        ),
    )


def test_call_revoke_works_with_dict():
    class Provider(provider.Provider):
        pass

    @access.target()
    class ExampleTarget:
        pass

    got_var = None

    @access.revoke()
    def revoke(p: Provider, subject: str, target: ExampleTarget, state: dict):
        nonlocal got_var
        got_var = state["test"]

    access.call_access_func(
        type="revoke",
        p=Provider(),
        data=rpc.GrantData(
            subject="test",
            target=rpc.GrantData.Target(
                kind="ExampleTarget",
                arguments={},
            ),
            state={"test": "example"},
        ),
    )

    assert got_var == "example"


def test_call_revoke_passes_request_id():
    class Provider(provider.Provider):
        pass

    @access.target()
    class ExampleTarget:
        pass

    got_var = None

    @access.revoke()
    def revoke(
        p: Provider, subject: str, target: ExampleTarget, request: rpc.AccessRequest
    ):
        nonlocal got_var
        got_var = request.id

    access.call_access_func(
        type="revoke",
        p=Provider(),
        data=rpc.GrantData(
            subject="test",
            target=rpc.GrantData.Target(
                kind="ExampleTarget",
                arguments={},
            ),
            request=rpc.AccessRequest(id="req_123"),
        ),
    )

    assert got_var == "req_123"
